#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <arpa/inet.h>


#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include "server.h"
#include "connections.h"
#include "mod_status.h"
#include "response.h"
#include "connections.h"
#include "log.h"

#include "plugin.h"


typedef struct {
	PLUGIN_DATA;
	
	double traffic_out;
	double requests;
	
	double mod_5s_traffic_out[5];
	double mod_5s_requests[5];
	size_t mod_5s_ndx;
	
	double rel_traffic_out;
	double rel_requests;
	
	double abs_traffic_out;
	double abs_requests;
	
	double bytes_written;
	
	buffer *rrd;
	buffer *module_list;
	
	int rrd_reports;
	buffer *rrd_dir;
} plugin_data;

INIT_FUNC(mod_status_init) {
	plugin_data *p;
	size_t i;
	
	p = calloc(1, sizeof(*p));
	
	p->traffic_out = p->requests = 0;
	p->rel_traffic_out = p->rel_requests = 0;
	p->abs_traffic_out = p->abs_requests = 0;
	p->bytes_written = 0;
	p->rrd = buffer_init();
	p->module_list = buffer_init();
	p->rrd_dir = buffer_init();
	
	for (i = 0; i < 5; i++) {
		p->mod_5s_traffic_out[i] = p->mod_5s_requests[i] = 0;
	}
	
	return p;
}

FREE_FUNC(mod_status_free) {
	plugin_data *p = p_d;
	
	UNUSED(srv);

	if (!p) return HANDLER_GO_ON;
	
	buffer_free(p->rrd);
	buffer_free(p->module_list);
	buffer_free(p->rrd_dir);
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_status_set_defaults) {
	plugin_data *p = p_d;
	
	config_values_t cv[] = { 
		{ "status.rrd-reports",          NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },
		{ "status.rrd-dir",              NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
		{ NULL,                          NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
	};
	
	if (!p) return HANDLER_ERROR;
	cv[0].destination = &(p->rrd_reports);
	cv[1].destination = p->rrd_dir;
	
	if (0 != config_insert_values(srv, cv)) {
		return HANDLER_ERROR;
	}
	
	/* check for dir */
	
	if (p->rrd_reports) {
		struct stat st;
		if (p->rrd_dir->used == 0) {
			log_error_write(srv, __FILE__, __LINE__, "s", "status.rrd-dir has to be set if status.rrd-reports is enabled");
		
			return HANDLER_ERROR;
		}
	
		if (0 != stat(p->rrd_dir->ptr, &st)) {
			log_error_write(srv, __FILE__, __LINE__, "sbs", "can't stat status.rrd-dir", p->rrd_dir, strerror(errno));
		
			return HANDLER_ERROR;
		}
	}
	
	return HANDLER_GO_ON;
}



static int mod_status_row_append(buffer *b, const char *key, const char *value) {
	buffer_append_string(b, "   <tr>\n");
	buffer_append_string(b, "    <td><b>");
	buffer_append_string(b, key);
	buffer_append_string(b, "</b></td>\n");
	buffer_append_string(b, "    <td>");
	buffer_append_string(b, value);
	buffer_append_string(b, "</td>\n");
	buffer_append_string(b, "   </tr>\n");
	
	return 0;
}

static int mod_status_header_append(buffer *b, const char *key) {
	buffer_append_string(b, "   <tr>\n");
	buffer_append_string(b, "    <th colspan=\"2\">");
	buffer_append_string(b, key);
	buffer_append_string(b, "</th>\n");
	buffer_append_string(b, "   </tr>\n");
	
	return 0;
}

static handler_t mod_status_handle_server_status(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	buffer *b;
	size_t j;
	double avg;
	
	b = chunkqueue_get_append_buffer(con->write_queue);
	
	buffer_copy_string(b, 
		      "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
		      "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
		      "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
		      "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
		      " <head>\n"
		      "  <title>Status</title>\n"
		      " </head>\n"
		      " <body>\n");
	
	/* connection listing */
	buffer_append_string(b, "<pre>");
	
	buffer_append_string(b, "<b>absolute</b> (since start)\n");
	
	buffer_append_string(b, " <b>Requests</b>: ");
	buffer_append_long(b, p->abs_requests);
	buffer_append_string(b, " req\n");
	
	buffer_append_string(b, " <b>Traffic</b> : ");
	buffer_append_long(b, p->abs_traffic_out / 1024);
	buffer_append_string(b, " kbytes\n");
	
	
	buffer_append_string(b, "<b>average</b> (5s sliding average)\n");
	for (j = 0, avg = 0; j < 5; j++) {
		avg += p->mod_5s_requests[j];
	}
	
	avg /= 5;
	
	buffer_append_string(b, " <b>Requests</b>: ");
	buffer_append_long(b, avg);
	buffer_append_string(b, " req/s\n");
	
	for (j = 0, avg = 0; j < 5; j++) {
		avg += p->mod_5s_traffic_out[j];
	}
	
	avg /= 5;
	
	buffer_append_string(b, " <b>Traffic</b> : ");
	buffer_append_long(b, avg / 1024);
	buffer_append_string(b, " kb/s\n");
	
	
	buffer_append_string(b, "\n<b>legend</b>\n");
	buffer_append_string(b, ". = connect, C = close, E = hard error\n");
	buffer_append_string(b, "r = read, R = read-POST, W = write, h = handle-request\n");
	buffer_append_string(b, "q = request-start,  Q = request-end\n");
	buffer_append_string(b, "s = response-start, S = response-end\n<hr />\n");
	
	for (j = 0; j < srv->conns->used; j++) {
		connection *c = srv->conns->ptr[j];
		const char *state = connection_get_short_state(c->state);
		
		if (c->fd == srv->fd) {
			buffer_append_string_len(b, " ", 1);
		} else {
			buffer_append_string_len(b, state, 1);
		}
		
		if (((j + 1) % 50) == 0) {
			buffer_append_string(b, "\n");
		}
	}
	
	buffer_append_string(b, "\n<hr />\n<b>Connections</b>\n");
	
	for (j = 0; j < srv->conns->used; j++) {
		connection *c = srv->conns->ptr[j];
		int ndx, i;
		
		/* ignore server connection */
		if (c->fd == srv->fd) continue;
		
#ifdef HAVE_IPV6	
		ndx = 0;
		for (i = 0; i < INET_NTOP_CACHE_MAX; i++) {
			if (srv->inet_ntop_cache[i].ts != 0) {
				if (srv->inet_ntop_cache[i].family == AF_INET6 &&
				    0 == memcmp(srv->inet_ntop_cache[i].addr.ipv6.s6_addr, c->dst_addr.ipv6.sin6_addr.s6_addr, 16)) {
					/* IPv6 found in cache */
					break;
				} else if (srv->inet_ntop_cache[i].family == AF_INET &&
					   srv->inet_ntop_cache[i].addr.ipv4.s_addr == c->dst_addr.ipv4.sin_addr.s_addr) {
					/* IPv4 found in cache */
					break;
					
				}
			}
		}
		
		if (i == INET_NTOP_CACHE_MAX) {
			/* not found in cache */
			
			i = ndx;
			inet_ntop(c->dst_addr.plain.sa_family, 
				  c->dst_addr.plain.sa_family == AF_INET6 ? 
				  (const void *) &(c->dst_addr.ipv6.sin6_addr) :
				  (const void *) &(c->dst_addr.ipv4.sin_addr),
				  srv->inet_ntop_cache[i].b2, INET6_ADDRSTRLEN);
			
			srv->inet_ntop_cache[i].ts = srv->cur_ts;
			srv->inet_ntop_cache[i].family = c->dst_addr.plain.sa_family;
			
			if (srv->inet_ntop_cache[i].family == AF_INET) {
				srv->inet_ntop_cache[i].addr.ipv4.s_addr = c->dst_addr.ipv4.sin_addr.s_addr;
			} else if (srv->inet_ntop_cache[i].family == AF_INET6) {
				memcpy(srv->inet_ntop_cache[i].addr.ipv6.s6_addr, c->dst_addr.ipv6.sin6_addr.s6_addr, 16);
			}
		}
#endif
	
#ifdef HAVE_IPV6
		buffer_append_string(b, srv->inet_ntop_cache[i].b2);
#else
		buffer_append_string(b, inet_ntoa(c->dst_addr.ipv4.sin_addr));
#endif
		
		buffer_append_string(b, " ");
		
		buffer_append_off_t(b, chunkqueue_written(c->write_queue));
		
		buffer_append_string(b, " ");
		
		buffer_append_string(b, connection_get_state(c->state));
		
		buffer_append_string(b, " ");
		
		buffer_append_string_buffer(b, c->physical.path);
		
		buffer_append_string(b, "\n");
	}
	
	
	buffer_append_string(b, 
		      "</pre>\n");
	
	
	buffer_append_string(b, 
		      " </body>\n"
		      "</html>\n"
		      );
	
	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
	
	con->http_status = 200;
	con->file_finished = 1;
	
	return HANDLER_FINISHED;
}

static handler_t mod_status_handle_server_config(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	buffer *b, *m = p->module_list;
	size_t i;
	
	struct ev_map { fdevent_handler_t et; const char *name; } event_handlers[] = 
	{ 
		/* - poll is most reliable
		 * - select works everywhere
		 * - linux-* are experimental
		 */
#ifdef USE_POLL
		{ FDEVENT_HANDLER_POLL,           "poll" },
#endif
#ifdef USE_SELECT
		{ FDEVENT_HANDLER_SELECT,         "select" },
#endif
#ifdef USE_LINUX_EPOLL
		{ FDEVENT_HANDLER_LINUX_SYSEPOLL, "linux-sysepoll" },
#endif
#ifdef USE_LINUX_SIGIO
		{ FDEVENT_HANDLER_LINUX_RTSIG,    "linux-rtsig" },
#endif
#ifdef USE_SOLARIS_DEVPOLL
		{ FDEVENT_HANDLER_SOLARIS_DEVPOLL,"solaris-devpoll" },
#endif
#ifdef USE_FREEBSD_KQUEUE
		{ FDEVENT_HANDLER_FREEBSD_KQUEUE, "freebsd-kqueue" },
#endif
		{ FDEVENT_HANDLER_UNSET,          NULL }
	};
	
	b = chunkqueue_get_append_buffer(con->write_queue);
	
	buffer_copy_string(b, 
			   "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
			   "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
			   "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
			   "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
			   " <head>\n"
			   "  <title>Status</title>\n"
			   " </head>\n"
			   " <body>\n"
			   "  <h1>"PACKAGE_NAME" "PACKAGE_VERSION"</h1>\n"
			   "  <table border=\"1\">\n");
	
	mod_status_header_append(b, "Server-Features");
#ifdef HAVE_PCRE_H
	mod_status_row_append(b, "Rewrite Engine", "enabled");
#else
	mod_status_row_append(b, "Rewrite Engine", "disabled - pcre missing");
#endif
#ifdef HAVE_ZLIB_H
	mod_status_row_append(b, "On-the-Fly Output Compression", "enabled");
#else
	mod_status_row_append(b, "On-the-Fly Output Compression", "disabled - zlib missing");
#endif
	mod_status_header_append(b, "Network Engine");
	
	for (i = 0; event_handlers[i].name; i++) {
		if (event_handlers[i].et == srv->event_handler) {
			mod_status_row_append(b, "fd-Event-Handler", event_handlers[i].name);
			break;
		}
	}
	
	mod_status_header_append(b, "Config-File-Settings");
	mod_status_row_append(b, "Directory Listings", con->conf.dir_listing ? "enabled" : "disabled");
	
	for (i = 0; i < srv->plugins.used; i++) {
		plugin **ps = srv->plugins.ptr;
		
		plugin *pl = ps[i];
	
		if (i == 0) {
			buffer_copy_string_buffer(m, pl->name);
		} else {
			buffer_append_string(m, "<br />");
			buffer_append_string_buffer(m, pl->name);
		}
	}
	
	mod_status_row_append(b, "Loaded Modules", m->ptr);
	
	buffer_append_string(b, "  </table>\n");
	
	buffer_append_string(b, 
		      " </body>\n"
		      "</html>\n"
		      );
	
	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
	
	con->http_status = 200;
	con->file_finished = 1;
	
	return HANDLER_FINISHED;
}

static handler_t mod_status_handler(server *srv, connection *con, void *p_d) {
	if (0 == strcmp(con->uri.path->ptr, "/server-status")) {
		return mod_status_handle_server_status(srv, con, p_d);
	} else if (0 == strcmp(con->uri.path->ptr, "/server-config")) {
		return mod_status_handle_server_config(srv, con, p_d);
	}
	
	return HANDLER_GO_ON;
}

TRIGGER_FUNC(mod_status_trigger) {
	plugin_data *p = p_d;
	int timeframe = 10;
	size_t i;
	
	/* check all connections */
	for (i = 0; i < srv->conns->used; i++) {
		connection *c = srv->conns->ptr[i];
		
		p->bytes_written += c->bytes_written - c->bytes_written_last_sec;
		c->bytes_written_last_sec = c->bytes_written;
	}
	
	/* a sliding average */
	p->mod_5s_traffic_out[p->mod_5s_ndx] = p->bytes_written;
	p->mod_5s_requests   [p->mod_5s_ndx] = p->requests;
	
	p->mod_5s_ndx = (p->mod_5s_ndx+1) % 5;
	
	p->abs_traffic_out += p->bytes_written;
	p->rel_traffic_out += p->bytes_written;
	
	p->bytes_written = 0;
	
	/* write the data down every 5 minutes */
	if ((p->rrd_reports) && 
	    ((srv->cur_ts % timeframe) == 0)) {
		int fd;
		double avg;
		char buf[16];
		
		strftime(buf, sizeof(buf), "%Y%m%d", gmtime(&(srv->cur_ts)));
		
		buffer_reset(p->rrd);
		buffer_copy_string_buffer(p->rrd, p->rrd_dir);
		BUFFER_APPEND_SLASH(p->rrd);
		buffer_append_string(p->rrd, "status.data-");
		buffer_append_string(p->rrd, buf);
		
		if (-1 != (fd = open(p->rrd->ptr, O_APPEND | O_WRONLY | O_CREAT | O_LARGEFILE, 0644))) {
			buffer_reset(p->rrd);
		
			buffer_copy_long(p->rrd, srv->cur_ts);
			buffer_append_string(p->rrd, ":");
			
			avg = (p->rel_traffic_out / timeframe);
			
			buffer_append_long(p->rrd, avg);
			buffer_append_string(p->rrd, ":");
			
			avg = (p->rel_requests / timeframe);
			
			buffer_append_long(p->rrd, avg);
			buffer_append_string(p->rrd, "\n");
			
			write(fd, p->rrd->ptr, p->rrd->used - 1);
			close(fd);
		} else {
			log_error_write(srv, __FILE__, __LINE__, "sbs", 
					"[mod_status] rrdfile:", 
					p->rrd,
					strerror(errno));
		}
		
		p->rel_requests = 0;
		p->rel_traffic_out = 0;
	}
	
	/* reset storage - second */
	p->traffic_out = 0;
	p->requests    = 0;
	
	return HANDLER_GO_ON;
}

REQUESTDONE_FUNC(mod_status_account) {
	plugin_data *p = p_d;
	
	UNUSED(srv);

	p->requests++;
	p->rel_requests++;
	p->abs_requests++;
	
	p->bytes_written += con->bytes_written;
	
	return HANDLER_GO_ON;
}

int mod_status_plugin_init(plugin *p) {
	p->name        = buffer_init_string("status");
	
	p->init        = mod_status_init;
	p->cleanup     = mod_status_free;
	p->set_defaults= mod_status_set_defaults;
	
	p->handle_uri_clean    = mod_status_handler;
	p->handle_trigger      = mod_status_trigger;
	p->handle_request_done = mod_status_account;
	
	p->data        = NULL;
	
	return 0;
}
